/**
 * Copyright 2012 Julien Eluard
 * This project includes software developed by Julien Eluard: https://github.com/jeluard/
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.github.jeluard.guayaba.util.concurrent;

import com.github.jeluard.guayaba.lang.Cancelable;
import com.github.jeluard.guayaba.lang.Runnables;
import com.github.jeluard.guayaba.lang.UncaughtExceptionHandlers;
import com.google.common.base.Preconditions;
import com.google.common.util.concurrent.ThreadFactoryBuilder;

import java.util.concurrent.*;
import java.util.logging.Logger;

/**
 * Helper methods for {@link ScheduledExecutorService}.
 */
public final class ScheduledExecutorServices {

  private ScheduledExecutorServices() {
  }

  /**
   * Schedule periodic execution of specified {@link Runnable}.
   * <br />
   * Each call to this method will create a new {@link Thread}.
   * <br />
   * {@link Exception} thrown by {@link Runnable#run()} will be intercepted and logged using `logger`.
   *
   * @param nameFormat 
   * @param runnable
   * @param initialDelay
   * @param period
   * @param unit
   * @param logger 
   * @return a {@link Cancelable} allowing to stop periodic execution.
   * @see Executors#newSingleThreadScheduledExecutor(java.util.concurrent.ThreadFactory) 
   */
  public static Cancelable scheduleWithFixedDelay(final String nameFormat, final Runnable runnable, final long initialDelay, final long period, final TimeUnit unit, final Logger logger) {
    Preconditions.checkNotNull(nameFormat, "null nameFormat");
    Preconditions.checkNotNull(runnable, "null runnable");
    Preconditions.checkArgument(initialDelay >= 0, "negative initialDelay");
    Preconditions.checkArgument(period > 0, "negative period");
    Preconditions.checkNotNull(unit, "null unit");
    Preconditions.checkNotNull(logger, "null logger");

    final ThreadFactory threadFactory = new ThreadFactoryBuilder().setDaemon(true).setPriority(Thread.MIN_PRIORITY).setNameFormat(nameFormat)
            .setThreadFactory(Executors.defaultThreadFactory()).setUncaughtExceptionHandler(UncaughtExceptionHandlers.defaultHandler(logger)).build();
    final ScheduledExecutorService scheduledExecutorService = Executors.newSingleThreadScheduledExecutor(threadFactory);
    final ScheduledFuture<?> scheduledFuture = scheduledExecutorService.scheduleWithFixedDelay(Runnables.loggingExecutionFailure(runnable, logger), initialDelay, period, unit);
    return new Cancelable() {
      @Override
      public void cancel() {
        scheduledFuture.cancel(true);//First cancel the task
        ExecutorServices.shutdownAndAwaitTermination(scheduledExecutorService, period, unit, logger);//Then force ScheduledExecutorService termination.
      }
    };
  }

}