/**
 * Copyright 2012 Julien Eluard
 * This project includes software developed by Julien Eluard: https://github.com/jeluard/
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.github.jeluard.guayaba.util.concurrent;

import com.github.jeluard.guayaba.lang.Cancelable;
import com.github.jeluard.guayaba.lang.Runnables;
import com.github.jeluard.guayaba.lang.UncaughtExceptionHandlers;
import com.google.common.base.Preconditions;
import com.google.common.base.Predicate;
import com.google.common.collect.Collections2;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import com.google.common.util.concurrent.Uninterruptibles;

import java.util.Collection;
import java.util.List;
import java.util.concurrent.*;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Simple scheduler allowing periodic execution of {@link Runnable}.
 * <br />
 * One {@link Thread} per `period`/`unit` will be created. Then a shared {@link ExecutorService} is used to execute all {@link Runnable tasks}.
 */
public final class Scheduler implements Cancelable {

  /**
   * Handle scheduled execution of a {@link List} of {@link Runnable} with the same period.
   */
  private static final class ScheduledRunnables implements Cancelable {

    private static final String TRIGGER_THREAD_PREFIX = "Scheduler Trigger - ";
    private final Cancelable thread;
    private final List<Callable<Void>> tasks = new CopyOnWriteArrayList<Callable<Void>>();

    private ScheduledRunnables(final long period, final TimeUnit unit, final ExecutorService executorService, final Logger logger) {
      final long maxAllowedExecutionTimeMillis = Math.max(unit.toMillis(period) / 2, 1);
      this.thread = ScheduledExecutorServices.scheduleWithFixedDelay(triggerThreadName(period, unit), new Runnable() {
        @Override
        public void run() {
          try {
            if (logger.isLoggable(Level.FINE)) {
              logger.log(Level.FINE, "About to execute <{0}>", ScheduledRunnables.this.tasks);
            }

            //TODO Investigate other option. Lots of useless Callable/FutureTask creation. See AbstractExecutorService for details.
            final List<Future<Void>> results = executorService.invokeAll(ScheduledRunnables.this.tasks, maxAllowedExecutionTimeMillis, TimeUnit.MILLISECONDS);

            if (logger.isLoggable(Level.FINE)) {
              logger.log(Level.FINE, "Got results <{0}>", results);
            }

            if (logger.isLoggable(Level.WARNING)) {
              //Tasks not successfully executed during allowed time will be returned as cancelled.
              final Collection<Future<Void>> resultsNotCompleted = Collections2.filter(results, new Predicate<Future<?>>() {
                @Override
                public boolean apply(final Future<?> future) {
                  return future.isCancelled();
                }
              });
              if (!resultsNotCompleted.isEmpty()) {
                //TODO map to original Future to propagate toString
                logger.log(Level.WARNING, "Failed to execute <{0}> in <{1}> ms.", new Object[]{resultsNotCompleted, maxAllowedExecutionTimeMillis});
              }
            }
          } catch (InterruptedException e) {
            //Executing thread has been interrupted externally. Fail fast this execution.
            if (logger.isLoggable(Level.WARNING)) {
              logger.warning("Scheduled execution interrupted.");
            }

            Thread.currentThread().interrupt();
          } catch (RejectedExecutionException e) {
            //Could not invoke all Tasks. Underlying ThreadPool rejected it. Only option is to remove some tasks.
            if (logger.isLoggable(Level.WARNING)) {
              logger.log(Level.WARNING, "Failed to schedule all tasks. "+ExecutorService.class.getSimpleName()+" rejected execution.", e);
            }
          }
        }
      }, period, period, unit, logger);
    }

    private static String triggerThreadName(final long period, final TimeUnit unit) {
      return ScheduledRunnables.TRIGGER_THREAD_PREFIX+period+"["+unit+"]";
    }

    /**
     * Add a {@link Callable} to the {@link List} of currently scheduled ones.
     * @param callable 
     */
    public void add(final Callable<Void> callable) {
      this.tasks.add(callable);
    }

    /**
     * Remove an existing {@link Callable} to the {@link List} of currently scheduled ones.
     * @param callable 
     */
    public void remove(final Callable<Void> callable) {
      this.tasks.remove(callable);
    }

    /**
     * Clear all currently scheduled {@link Callable}.
     */
    public void clear() {
      this.tasks.clear();
    }

    /**
     * Cancel scheduling process.
     */
    @Override
    public void cancel() {
      clear();
      this.thread.cancel();
    }

  }

  private static final String THREAD_EXECUTOR_NAME = "Scheduler Executor #%s";
  private final ScheduledRunnables scheduledRunnables;

  /**
   * @param period
   * @param unit
   * @param executorService 
   */
  public Scheduler(final long period, final TimeUnit unit, final ExecutorService executorService, final Logger logger) {
    Preconditions.checkNotNull(unit, "null unit");
    Preconditions.checkNotNull(executorService, "null executorService");
    Preconditions.checkNotNull(logger, "null logger");

    this.scheduledRunnables = new ScheduledRunnables(period, unit, executorService, logger);
  }

  /**
   * @param maxTasks
   * @return a default {@link ExecutorService}. Assumes it will be used with {@link Scheduler} only.
   */
  public static ExecutorService defaultExecutorService(final int maxTasks, final Logger logger) {
    final int availableProcessors = Runtime.getRuntime().availableProcessors();
    final ThreadFactory threadFactory = new ThreadFactoryBuilder().setDaemon(true).setPriority(Thread.MIN_PRIORITY).setNameFormat(Scheduler.THREAD_EXECUTOR_NAME)
            .setThreadFactory(Executors.defaultThreadFactory()).setUncaughtExceptionHandler(UncaughtExceptionHandlers.defaultHandler(logger)).build();
    final ExecutorService executorService = new ThreadPoolExecutor(availableProcessors, availableProcessors*2, 1l, TimeUnit.MINUTES, new ArrayBlockingQueue<Runnable>(maxTasks, true), threadFactory, new ThreadPoolExecutor.AbortPolicy()) {
      @Override
      protected void beforeExecute(final Thread thread, final Runnable runnable) {
        super.beforeExecute(thread, runnable);

        if (logger.isLoggable(Level.FINE)) {
          logger.log(Level.FINE, "About to execute <{0}> from <{1}>", new Object[]{runnable, thread});
        }
      }
      @Override
      protected void afterExecute(final Runnable runnable, final Throwable t) {
        super.afterExecute(runnable, t);

        //No Runnable are directly executed in this ThreadPool so `t` will always be null.

        //Logs success/failures when applicable.
        if (logger.isLoggable(Level.WARNING)) {
          final Future future = Future.class.cast(runnable); //Safe cast. We never execute Runnable directly.
          if (future.isCancelled()) {//Only consider successful executions, ignore cancelled tasks.
            if (logger.isLoggable(Level.FINE)) {
              logger.log(Level.FINE, "Successfully executed <{0}>", runnable);
            }

            return;
          }

          try {
            Uninterruptibles.getUninterruptibly(future);
          } catch (ExecutionException e) {
            logger.log(Level.WARNING, "Failed to execute <"+runnable+">", e.getCause());
          }
        }
      }
    };
    return executorService;
  }

  /**
   * Schedule specified {@link Runnable} for execution every `period`.
   *
   * @param runnable
   * @return a {@link Cancelable} allowing to cancel scheduling of specified {@link Runnable}
   */
  public Cancelable schedule(final Runnable runnable) {
    Preconditions.checkNotNull(runnable, "null runnable");

    final Callable<Void> callable = Runnables.asCallable(runnable);
    this.scheduledRunnables.add(callable);

    return new Cancelable() {
      @Override
      public void cancel() {
        Scheduler.this.scheduledRunnables.remove(callable);
      }
    };
  }

  /**
   * Clear all currently scheduled {@link Runnable}. Current execution (if any) will be finished.
   */
  public void clear() {
    this.scheduledRunnables.clear();
  }

  /**
   * Cancel all scheduled {@link Runnable} and background scheduler {@link Thread}.
   */
  @Override
  public void cancel() {
    this.scheduledRunnables.cancel();
  }

}